@@ -185,8 +185,11 @@ HTML_FOOTER = """
  <div id="footer">
   <div>
    piuparts was written by <a href="mailto:liw@iki.fi">Lars Wirzenius</a> and is now maintained by 
-   <a href="mailto:holger@debian.org">Holger Levsen</a>,  
-   <a href="mailto:luk@debian.org">Luk Claes</a> and <a href="http://qa.debian.org/" target="_blank">others</a>. 
+   <a href="mailto:holger@debian.org">Holger Levsen</a> and
+   <a href="mailto:piuparts-devel@lists.alioth.debian.org">others</a> using
+   <a href="http://anonscm.debian.org/gitweb/?p=piuparts/piuparts.git" target="_blank">piuparts.git</a> and
+   <a href="http://anonscm.debian.org/gitweb/?p=piuparts/piatti.git" target="_blank">piatti.git</a> and the
+   <a href="http://packages.qa.debian.org/piuparts" target="_blank">PTS</a>.
    GPL2 <a href="http://packages.debian.org/changelogs/pool/main/p/piuparts/current/copyright" target="_blank">licenced</a>.
    Weather icons are from the <a href="http://tango.freedesktop.org/Tango_Icon_Library" target="_blank">Tango Icon Library</a>.
    <a href="http://validator.w3.org/check?uri=referer">
@@ -366,6 +369,7 @@ linktarget_by_template = {
     "owned_files_after_purge_error.tpl": "due to owned files existing after purge",
     "owned_files_by_many_packages_error.tpl": "due to owned files by many packages",
     "processes_running_error.tpl": "due to leaving processes running behind",
+    "excessive_output_error.tpl": "due to being terminated after excessive output",
     "unowned_files_after_purge_error.tpl": "due to unowned files after purge",
     "conffile_prompt_error.tpl": "due to prompting due to modified conffiles",
     "unknown_purge_error.tpl": "due to purge failed due to an unknown reason",
@@ -386,6 +390,7 @@ linktarget_by_template = {
     "modified_files_after_purge_error.tpl": "due to files having been modified after purge",
     "disappeared_files_after_purge_error.tpl": "due to files having disappeared after purge",
     "db_setup_error.tpl": "due to failing to setup a database",
+    "dependency_error.tpl": "due to unsatisfied dependencies",
 }
 
 
@@ -415,7 +420,7 @@ def setup_logging(log_level, log_file_name):
     handler = logging.StreamHandler(sys.stderr)
     handler.setFormatter(formatter)
     logger.addHandler(handler)
-    
+
     if log_file_name:
         handler = logging.FileHandler(log_file_name)
         logger.addHandler(handler)
@@ -431,7 +436,23 @@ def html_protect(vstr):
 
 
 def emphasize_reason(reason):
-    if reason in ("unknown", "failed-testing", "circular-dependency", "dependency-failed-testing", "dependency-does-not-exist", "cannot-be-tested"):
+    bad_states = [
+        #"successfully-tested",
+        "failed-testing",
+        "cannot-be-tested",
+        #"essential-required",
+        #"waiting-to-be-tested",
+        #"waiting-for-dependency-to-be-tested",
+        "dependency-failed-testing",
+        "dependency-cannot-be-tested",
+        "dependency-does-not-exist",
+        "circular-dependency",
+        "does-not-exist",
+        "unknown",
+        "unknown-preferred-alternative",
+        "no-dependency-from-alternatives-exists",
+    ]
+    if reason in bad_states:
       reason = "<em>"+reason+"</em>"
     return reason
 
@@ -625,7 +646,7 @@ class Section:
         return links
 
     def link_to_maintainer_summary(self, maintainer):
-	email = get_email_address(maintainer)
+        email = get_email_address(maintainer)
         return "<a href=\"/%s/maintainer/%s/%s.html\">%s</a>" % (self._config.section,maintainer_subdir(email),email,html_protect(maintainer))
 
     def link_to_uploaders(self, uploaders):
@@ -643,8 +664,8 @@ class Section:
         return link
 
     def link_to_state_page(self, section, package_name, link_target):
-        state = self._binary_db.state_by_name(package_name)
-        if state != "unknown":
+        if self._binary_db.has_package(package_name):
+            state = self._binary_db.get_package_state(package_name)
             link = "<a href=\"/%s/%s\">%s</a>" % (
                 section,
                 "state-"+state+".html"+"#"+package_name,
@@ -684,11 +705,13 @@ class Section:
         counts = current_day
         total = 0
         for state in self._binary_db.get_states():
-            count = len(self._binary_db.get_packages_in_state(state))
+            count = len(self._binary_db.get_pkg_names_in_state(state))
             header += ", %s" % state
             counts += ", %s" % count
             logging.debug("%s: %s" % (state, count))
             total += count
+        logging.debug("total: %s" % total)
+        logging.debug("source: %s" % len(self._source_db.get_all_packages()))
         header += "\n"       
         counts += "\n"       
 
@@ -705,33 +728,36 @@ class Section:
         return total
 
 
-    def merge_maintainer_templates(self, templates):
-        for maint_tpl in templates:
-            tpl = os.path.join(self._output_directory,"maintainer",maintainer_subdir(maint_tpl),maint_tpl)
-            lines = read_file(tpl)
+    def create_maintainer_summaries(self, maintainers, source_data):
+        logging.debug("Writing maintainer summaries in %s" % self._output_directory)
+        maintainer_dir = os.path.join(self._output_directory, "maintainer")
+        if not os.path.exists(maintainer_dir):
+            os.mkdir(maintainer_dir)
+        states = ["fail", "unknown", "pass"]
+        for maintainer in maintainers.keys():
+            sources = maintainers[maintainer]
+            maintainer_subdir_path = os.path.join(maintainer_dir, maintainer_subdir(maintainer))
+            if not os.path.exists(maintainer_subdir_path):
+              os.mkdir(maintainer_subdir_path)
             rows = ""
-            for line in lines:
-              state, count, packages = line.split(",")
-              if packages == "none\n":
-                links = "&nbsp;"
-              else:
-                links = ""
-                for package in packages.split(" "):
-                  links += "<a href=\"#%s\">%s</a> " % (package,package)
-              rows += "<tr class=\"normalrow\"><td class=\"labelcell\">%s:</td><td class=\"contentcell2\">%s</td><td class=\"contentcell2\" colspan=\"4\">%s</td></tr>" % \
-                       (state, count, links)
-            os.unlink(tpl)
-            template_path = tpl[:-len("_tpl")]
+            package_rows = ""
+            packages = {}
+            for state in states:
+                packages[state] = []
+            for source in sorted(sources):
+                (state, sourcerows, binaryrows) = source_data[source]
+                packages[state].append(source)
+                package_rows += sourcerows + binaryrows
 
-            for state in ("fail","unkn","pass"):
-                filename = template_path+"_"+state
-                if os.path.isfile(filename):
-                     f = file(filename, "r")
-                     rows += file.read(f)
-                     f.close()
-                     os.unlink(filename)
-
-            maintainer = os.path.basename(maint_tpl[:-len("_tpl")])
+            for state in states:
+                if len(packages[state]) > 0:
+                    links = ""
+                    for package in packages[state]:
+                        links += "<a href=\"#%s\">%s</a> " % (package, package)
+                else:
+                    links = "&nbsp;"
+                rows += "<tr class=\"normalrow\"><td class=\"labelcell\">%s:</td><td class=\"contentcell2\">%s</td><td class=\"contentcell2\" colspan=\"4\">%s</td></tr>" % \
+                        (state, len(packages[state]), links)
 
             distrolinks = "<tr class=\"normalrow\"><td class=\"labelcell\">other distributions: </td><td class=\"contentcell2\" colspan=\"5\">"
             for section in self._section_names:
@@ -740,7 +766,7 @@ class Section:
             distrolinks += "</td></tr>"
 
             htmlpage = string.Template(HTML_HEADER + MAINTAINER_BODY_TEMPLATE + HTML_FOOTER)
-            filename = template_path+".html"
+            filename = os.path.join(maintainer_subdir_path, maintainer + ".html")
             f = file(filename, "w")
             f.write(htmlpage.safe_substitute( {
                "page_title": html_protect("Status of "+maintainer+" packages in "+self._config.section),
@@ -748,10 +774,11 @@ class Section:
                "distrolinks": distrolinks,
                "section_navigation": create_section_navigation(self._section_names,self._config.section),
                "time": time.strftime("%Y-%m-%d %H:%M %Z"),
-               "rows": rows,
+               "rows": rows + package_rows,
              }))
             f.close()
 
+
     def create_source_summary (self, source, logs_by_dir):
         source_version = self._source_db.get_control_header(source, "Version")
         binaries = self._source_db.get_control_header(source, "Binary")
@@ -761,8 +788,8 @@ class Section:
         success = True
         failed = False
         binaryrows = ""
-        for binary in binaries.split(", "):
-          state = self._binary_db.state_by_name(binary)
+        for binary in sorted(binaries.split(", ")):
+          state = self._binary_db.get_package_state(binary)
           if state == "unknown":
             # Don't track udebs and binary packages on other archs. 
             # The latter is a FIXME which needs parsing the Packages files from other archs too
@@ -790,7 +817,7 @@ class Section:
           sourcerows += "<tr class=\"normalrow\"><td class=\"labelcell\">Maintainer:</td><td class=\"contentcell2\" colspan=\"5\">%s</td></tr>" % (self.link_to_maintainer_summary(maintainer))
           if uploaders:
             sourcerows += "<tr class=\"normalrow\"><td class=\"labelcell\">Uploaders:</td><td class=\"contentcell2\" colspan=\"5\">%s</td></tr>" % (self.link_to_uploaders(uploaders))
-        
+
           source_summary_page_path = os.path.join(self._output_directory, "source", source_subdir(source))
           if not os.path.exists(source_summary_page_path):
              os.makedirs(source_summary_page_path)
@@ -814,61 +841,30 @@ class Section:
 
         return sourcerows, binaryrows, source_state, maintainer, uploaders
 
-    def create_maintainer_templates_for_source(self,source, source_state, sourcerows, binaryrows, maintainer, uploaders):
-        maintainer_pages = [] 
-        maintainer_pages.append(get_email_address(maintainer))
-        for uploader in uploaders.split(", "):
-          if uploader:
-            maintainer_pages.append(get_email_address(uploader))
-        for maintainer_page in maintainer_pages:
-          maintainer_summary_page_path = os.path.join(self._output_directory, "maintainer", maintainer_subdir(maintainer_page))
-
-          if not os.path.exists(maintainer_summary_page_path):
-            os.makedirs(maintainer_summary_page_path)
-          filename = os.path.join(maintainer_summary_page_path, (maintainer_page + "_tpl"))
-          maintainer_package_count = {}
-          maintainer_packages = {}
-          if os.path.isfile(filename):
-            lines = read_file(filename)
-            for line in lines:
-              state, count, packages = line.split(",")
-              maintainer_package_count[state]=int(count)
-              maintainer_packages[state]=packages[:-1]
-            if maintainer_packages[source_state] == "none":
-              maintainer_packages[source_state] = source
-            else:
-              maintainer_packages[source_state] = "%s %s" % (maintainer_packages[source_state],source)
-          else:
-            maintainer_package_count["fail"] = 0
-            maintainer_package_count["unknown"] = 0
-            maintainer_package_count["pass"] = 0
-            for state in "fail", "unknown", "pass":
-              maintainer_packages[state] = "none"
-            maintainer_packages[source_state] = source
-          if source_state == "fail":
-            maintainer_package_count["fail"]+=1
-          elif source_state == "unknown":
-            maintainer_package_count["unknown"]+=1
-          else:
-            maintainer_package_count["pass"]+=1
-          lines = ""
-          for state in "fail", "unknown", "pass":
-            lines +=  "%s,%s,%s\n" % (state,maintainer_package_count[state],maintainer_packages[state])
-          write_file(filename,lines)
-          append_file(filename[:-4]+"_"+source_state[:4],sourcerows+binaryrows)
 
     def create_package_summaries(self, logs_by_dir):
         logging.debug("Writing package templates in %s" % self._config.section)    
 
+        maintainers = {}
+        source_binary_rows = {}
         sources = ""
         for source in self._source_db.get_all_packages():
             (sourcerows, binaryrows, source_state, maintainer, uploaders) = self.create_source_summary(source, logs_by_dir)
             if source_state != "udeb":
-              sources += "%s: %s\n" % (source, source_state)
-              self.create_maintainer_templates_for_source(source, source_state, sourcerows, binaryrows, maintainer, uploaders)
- 
+                sources += "%s: %s\n" % (source, source_state)
+                source_binary_rows[source] = (source_state, sourcerows, binaryrows)
+                for maint in [maintainer] + uploaders.split(","):
+                    if maint.strip():
+                        email = get_email_address(maint.strip())
+                        if not "INVALID" in email:
+                            if not email in maintainers:
+                                maintainers[email] = []
+                            maintainers[email].append(source)
+
         write_file(os.path.join(self._output_directory, "sources.txt"), sources)
 
+        self.create_maintainer_summaries(maintainers, source_binary_rows)
+
 
     def make_stats_graph(self):
         countsfile = os.path.join(self._output_directory, "counts.txt")
@@ -952,7 +948,7 @@ class Section:
               analysis = self.create_and_link_to_analysises(state)
             tablerows += ("<tr class=\"normalrow\"><td class=\"contentcell2\"><a href='state-%s.html'>%s</a>%s</td>" +
                           "<td class=\"contentcell2\">%d</td><td class=\"contentcell2\">%s</td></tr>\n") % \
-                          (html_protect(state), html_protect(state), analysis, len(self._binary_db.get_packages_in_state(state)),
+                          (html_protect(state), html_protect(state), analysis, len(self._binary_db.get_pkg_names_in_state(state)),
                           dir_link)
         try:
           tablerows += self.make_stats_graph();
@@ -971,21 +967,45 @@ class Section:
             "packagesurl": html_protect(self._config["packages-url"]), 
            }))
 
+    def _show_providers(self, dep):
+        providers = self._binary_db.get_providers(dep)
+        vlist = ""
+        if providers:
+            vlist += "\n<ul>\n"
+            for provider in providers:
+                vlist += "<li>provider %s is %s</li>\n" % \
+                          (self.link_to_state_page(self._config.section, provider, provider),
+                          emphasize_reason(html_protect(self._binary_db.get_package_state(provider))))
+            vlist += "</ul>\n"
+        return vlist
+
     def write_state_pages(self):
         for state in self._binary_db.get_states():
             logging.debug("Writing page for %s" % state)
             vlist = ""
-            for package in self._binary_db.get_packages_in_state(state):
+            for name in sorted(self._binary_db.get_pkg_names_in_state(state)):
+                package = self._binary_db.get_package(name)
                 vlist += "<li id=\"%s\">%s (%s)" % (
                                          package["Package"],
                                          self.link_to_source_summary(package["Package"]),
                                          html_protect(package["Maintainer"]))
-                if package.dependencies():
+                all_deps = package.all_dependencies()
+                if all_deps:
                     vlist += "\n<ul>\n"
-                    for dep in package.dependencies():
+                    for alternatives in all_deps:
+                        dep = alternatives[0]
                         vlist += "<li>dependency %s is %s</li>\n" % \
                                   (self.link_to_state_page(self._config.section,dep,dep), 
-                                  emphasize_reason(html_protect(self._binary_db.state_by_name(dep))))
+                                  emphasize_reason(html_protect(self._binary_db.get_package_state(dep, resolve_virtual=False))))
+                        vlist += self._show_providers(dep)
+                        if len(alternatives) > 1:
+                            vlist += "\n<ul>\n"
+                            for dep in alternatives[1:]:
+                                vlist += "<li>alternative dependency %s is %s</li>\n" % \
+                                          (self.link_to_state_page(self._config.section,dep,dep),
+                                          emphasize_reason(html_protect(self._binary_db.get_package_state(dep, resolve_virtual=False))))
+                                vlist += self._show_providers(dep)
+                            vlist += "</ul>\n"
                     vlist += "</ul>\n"
                 vlist += "</li>\n"
             htmlpage = string.Template(HTML_HEADER + STATE_BODY_TEMPLATE + HTML_FOOTER)
@@ -998,6 +1018,25 @@ class Section:
                                         "list": vlist
                                        }))
 
+
+    def archive_logfile(self, vdir, log):
+        archivedir = os.path.join("archive", vdir)
+        if not os.path.exists(archivedir):
+            os.makedirs(archivedir)
+        os.rename(os.path.join(vdir, log), os.path.join("archive", vdir, log))
+
+
+    def cleanup_removed_packages(self, logs_by_dir):
+        for vdir in logs_by_dir.keys():
+            for log in sorted(logs_by_dir[vdir]):
+                if log.endswith(".log"):
+                    package, version = log[:-len(".log")].split("_")
+                    if not self._binary_db.has_package(package):
+                        logging.debug("Package %s was removed, archiving %s/%s" % (package, vdir, log))
+                        self.archive_logfile(vdir, log)
+                        logs_by_dir[vdir].remove(log)
+
+
     def generate_html(self):
         logging.debug("Finding log files")
         dirs = ["pass", "fail", "bugged", "reserved", "untestable"]
@@ -1005,6 +1044,9 @@ class Section:
         for vdir in dirs:
             logs_by_dir[vdir] = find_files_with_suffix(vdir, ".log")
 
+        logging.debug("Archiving logs of obsolete packages")
+        self.cleanup_removed_packages(logs_by_dir)
+
         logging.debug("Copying log files")
         copy_logs(logs_by_dir, self._output_directory)
 
@@ -1019,9 +1061,6 @@ class Section:
         if self._config["sources-url"]:
             self.create_package_summaries(logs_by_dir)
 
-            logging.debug("Merging maintainer summaries in %s" % self._output_directory)    
-            self.merge_maintainer_templates(find_files_with_suffix(self._output_directory+"/maintainer/", "_tpl"))
-
         logging.debug("Writing section index page")    
         self.write_section_index_page(dirs, total_packages)
 
@@ -1045,6 +1084,7 @@ class Section:
         self.generate_html()
         os.chdir(oldcwd)
 
+
 def main():
     setup_logging(logging.DEBUG, None)
 
@@ -1054,6 +1094,7 @@ def main():
     section_names = []
     if len(sys.argv) > 1:
         section = sys.argv[1]
+        # FIXME: does this work without settings master_directory and output_directory?
     else:
         global_config = Config(section="global")
         global_config.read(CONFIG_FILE)
@@ -1096,3 +1137,5 @@ def main():
 
 if __name__ == "__main__":
     main()
+
+# vi:set et ts=4 sw=4 :
