@@ -28,6 +28,8 @@ import sys
 import stat
 import time
 import logging
+from signal import alarm, signal, SIGALRM, SIGKILL
+import subprocess
 import ConfigParser
 
 import piupartslib.conf
@@ -35,7 +37,7 @@ import piupartslib.packagesdb
 
 
 CONFIG_FILE = "/etc/piuparts/piuparts.conf"
-
+MAX_WAIT_TEST_RUN = 45*60
 
 def setup_logging(log_level, log_file_name):
     logger = logging.getLogger()
@@ -47,7 +49,7 @@ def setup_logging(log_level, log_file_name):
     handler = logging.StreamHandler(sys.stderr)
     handler.setFormatter(formatter)
     logger.addHandler(handler)
-    
+
     if log_file_name:
         handler = logging.FileHandler(log_file_name)
         logger.addHandler(handler)
@@ -79,13 +81,20 @@ class Config(piupartslib.conf.Config):
                 "debug": "no",
                 "keep-sources-list": "no",
                 "arch": None,
+                "precedence": "1",
             }, "")
 
 
+class Alarm(Exception):
+    pass
+
+def alarm_handler(signum, frame):
+    raise Alarm
+
 class MasterNotOK(Exception):
 
     def __init__(self):
-        self.args = "Master did not responed with 'ok'"
+        self.args = "Master did not respond with 'ok'"
 
 
 class MasterDidNotGreet(Exception):
@@ -109,12 +118,12 @@ class Slave:
         self._master_user = None
         self._master_directory = "."
         self._master_command = None
-        
+
     def _readline(self):
         line = self._from_master.readline()
         logging.debug("<< " + line.rstrip())
         return line
-        
+
     def _writeline(self, *words):
         line = " ".join(words)
         logging.debug(">> " + line)
@@ -124,34 +133,32 @@ class Slave:
     def set_master_host(self, host):
         logging.debug("Setting master host to %s" % host)
         self._master_host = host
-        
+
     def set_master_user(self, user):
         logging.debug("Setting master user to %s" % user)
         self._master_user = user
-        
+
     def set_master_directory(self, dir):
         logging.debug("Setting master directory to %s" % dir)
         self._master_directory = dir
-        
+
     def set_master_command(self, cmd):
         logging.debug("Setting master command to %s" % cmd)
         self._master_command = cmd
-    
+
     def connect_to_master(self, log_file):
         logging.info("Connecting to %s" % self._master_host)
         if self._master_user:
-            user = "-l " + self._master_user
+            user = self._master_user + "@"
         else:
             user = ""
-        (self._to_master, self._from_master) = \
-            os.popen2("ssh %s %s 'cd %s; %s 2> %s.$$ && rm %s.$$'" %
-                                    (self._master_host,
-                                     user,
-                                     self._master_directory or ".",
-                                     self._master_command,
-                                     log_file,
-                                     log_file))
-
+        ssh_cmdline = "cd %s; %s 2> %s.$$ && rm %s.$$" % \
+                      (self._master_directory or ".", 
+                      self._master_command, log_file, log_file)
+        p = subprocess.Popen(["ssh", user + self._master_host, ssh_cmdline], 
+                       stdin=subprocess.PIPE, stdout=subprocess.PIPE)
+        self._to_master = p.stdin
+        self._from_master = p.stdout
         line = self._readline()
         if line != "hello\n":
             raise MasterDidNotGreet()
@@ -181,6 +188,15 @@ class Slave:
         if line != "ok\n":
             raise MasterNotOK()
 
+    def get_status(self):
+        self._writeline("status")
+        line = self._readline()
+        words = line.split()
+        if words and words[0] == "ok":
+            logging.info("Master status: " + " ".join(words[1:]))
+        else:
+            raise MasterIsCrazy()
+
     def reserve(self):
         self._writeline("reserve")
         line = self._readline()
@@ -236,16 +252,9 @@ class Section:
         if self._config["chroot-tgz"] and not self._config["distro"]:
           logging.info("The option --chroot-tgz needs --distro.")
 
-        tarball = self._config["chroot-tgz"]
-        if tarball:
-            create_or_replace_chroot_tgz(self._config, tarball,
-                      base_tgz_ctrl, self._config["distro"])
+        self._base_tgz_ctrl = base_tgz_ctrl
+        self._check_tarball()
 
-        tarball = self._config["upgrade-test-chroot-tgz"]
-        if self._config["upgrade-test-distros"] and tarball: 
-            create_or_replace_chroot_tgz(self._config, tarball,
-                      base_tgz_ctrl, self._config["upgrade-test-distros"].split()[0])
-    
         for rdir in ["new", "pass", "fail"]:
             rdir = os.path.join(self._slave_directory, rdir)
             if not os.path.exists(rdir):
@@ -264,9 +273,25 @@ class Section:
                 os.makedirs(rdir)
         os.chdir(oldcwd)
 
+    def _check_tarball(self):
+        tarball = self._config["chroot-tgz"]
+        if tarball:
+            create_or_replace_chroot_tgz(self._config, tarball,
+                      self._base_tgz_ctrl, self._config["distro"])
+
+        tarball = self._config["upgrade-test-chroot-tgz"]
+        if self._config["upgrade-test-distros"] and tarball:
+            create_or_replace_chroot_tgz(self._config, tarball,
+                      self._base_tgz_ctrl, self._config["upgrade-test-distros"].split()[0])
+
+    def precedence(self):
+        return int(self._config["precedence"])
+
     def run(self):
         logging.info("-------------------------------------------")
-        logging.info("Running section " + self._config.section)
+        logging.info("Running section %s (precedence=%d)" % (self._config.section, self.precedence()))
+        self._config = Config(section=self._config.section)
+        self._config.read(CONFIG_FILE)
         self._slave.connect_to_master(self._log_file)
 
         oldcwd = os.getcwd()
@@ -284,9 +309,13 @@ class Section:
             while len(self._slave.get_reserved()) < max_reserved and self._slave.reserve():
                 pass
 
+        self._slave.get_status()
         self._slave.close()
 
+        test_count = len(self._slave.get_reserved())
+
         if self._slave.get_reserved():
+            self._check_tarball()
             packages_files = {}
             if self._config["distro"]:
                 distros = [self._config["distro"]]
@@ -320,6 +349,8 @@ class Section:
                 self._slave.forget_reserved(package_name, version)
             os.chdir(oldcwd)
 
+        return( test_count )
+
 
 def log_name(package, version):
     return "%s_%s.log" % (package, version)
@@ -338,6 +369,41 @@ def upgrade_testable(config, package, packages_files):
     else:
         return False
 
+def get_process_children(pid):
+    p = subprocess.Popen('ps --no-headers -o pid --ppid %d' % pid,
+           shell = True, stdout = subprocess.PIPE, stderr = subprocess.PIPE)
+    stdout, stderr = p.communicate()
+    return [int(p) for p in stdout.split()]
+
+def run_test_with_timeout(cmd, maxwait, kill_all):
+      logging.debug("Executing: %s" % cmd)
+
+      stdout = ""
+      sh_cmd = "{ %s; } 2>&1" % cmd
+      p = subprocess.Popen(sh_cmd, shell=True,
+                     stdin=subprocess.PIPE, stdout=subprocess.PIPE)
+      if maxwait > 0:
+          signal(SIGALRM, alarm_handler)
+          alarm(maxwait)
+      try:
+          stdout, stderr = p.communicate()
+          if maxwait > 0:
+              alarm(0)
+      except Alarm:
+          pids = [p.pid]
+          if kill_all:
+              pids.extend(get_process_children(p.pid))
+          for pid in pids:
+              if pid > 0:
+                  try: 
+                      os.kill(pid, SIGKILL)
+                  except OSError:
+                      pass
+          return -1,stdout
+
+      return p.returncode,stdout 
+
+
 def test_package(config, package, packages_files):
     logging.info("Testing package %s/%s %s" % (config.section, package["Package"], package["Version"]))
 
@@ -350,56 +416,48 @@ def test_package(config, package, packages_files):
     output.write("\n")
     package.dump(output)
     output.write("\n")
-   
+
     # omit distro test if chroot-tgz is not specified.
+    ret = 0
     if config["chroot-tgz"]: 
-      command = "%(piuparts-cmd)s -ad %(distro)s -b %(chroot-tgz)s" % \
-                  config
-      if config["keep-sources-list"] in ["yes", "true"]:
-          command += " --keep-sources-list "
-    
-      if config["mirror"]:
-          command += " --mirror %s " % config["mirror"]
-      command += " " + package["Package"]
+        command = "%(piuparts-cmd)s -ad %(distro)s -b %(chroot-tgz)s" % \
+                    config
+        if config["keep-sources-list"] in ["yes", "true"]:
+            command += " --keep-sources-list "
+        if config["mirror"]:
+            command += " --mirror %s " % config["mirror"]
+        command += " " + package["Package"]
 
-      logging.debug("Executing: %s" % command)
-      output.write("Executing: %s\n" % command)
-      f = os.popen("{ %s; } 2>&1" % command, "r")
-      for line in f:
-          output.write(line)
-      status = f.close()
-      if status is None:
-          status = 0
-    else:
-          status = 0
+        output.write("Executing: %s\n" % command)
+        ret,f = run_test_with_timeout(command, MAX_WAIT_TEST_RUN, True)
+        if ret < 0:
+            output.write(f + "\n *** Process KILLED - exceed maximum run time ***\n")
+        else:
+            output.write(f)
 
-    if status == 0 and upgrade_testable(config, package, packages_files):
+    if ret == 0 and upgrade_testable(config, package, packages_files):
         distros = config["upgrade-test-distros"].split()
         distros = ["-d " + distro.strip() for distro in distros]
         distros = " ".join(distros)
         command = "%(piuparts-cmd)s -ab %(upgrade-test-chroot-tgz)s " % config
         command += distros
-
         if config["mirror"]:
           command += " --mirror %s " % config["mirror"]
-
         command += " " + package["Package"]
 
-        logging.debug("Executing: %s" % command)
-        output.write("\nExecuting: %s\n" % command)
-        f = os.popen("{ %s; } 2>&1" % command, "r")
-        for line in f:
-            output.write(line)
+        output.write("Executing: %s\n" % command)
+        ret,f = run_test_with_timeout(command, MAX_WAIT_TEST_RUN, True)
+        if ret < 0:
+            output.write(" *** Process KILLED - exceed maximum run time ***\n")
+        else:
+            output.write(f)
             output.flush()
-        status = f.close()
-        if status is None:
-            status = 0
 
     output.write("\n")
     output.write(time.strftime("End: %Y-%m-%d %H:%M:%S %Z\n", 
                                time.gmtime()))
     output.close()
-    if not os.WIFEXITED(status) or os.WEXITSTATUS(status) != 0:
+    if ret != 0:
         subdir = "fail"
     else:
         subdir = "pass"
@@ -408,14 +466,24 @@ def test_package(config, package, packages_files):
 
 
 def create_chroot(config, tarball, distro):
+    output_name = tarball + ".log"
+    logging.debug("Opening log file %s" % output_name)
     logging.info("Creating new tarball %s" % tarball)
     command = "%s -ad %s -s %s.new -m %s hello" % \
                 (config["piuparts-cmd"], distro, tarball, config["mirror"])
+    output = file(output_name, "w")
+    output.write(time.strftime("Start: %Y-%m-%d %H:%M:%S %Z\n\n",
+                               time.gmtime()))
+    output.write("Executing: " + command)
     logging.debug("Executing: " + command)
     f = os.popen("{ %s; } 2>&1" % command, "r")
     for line in f:
+        output.write(line)
         logging.debug(">> " + line.rstrip())
     f.close()
+    output.write(time.strftime("\nEnd: %Y-%m-%d %H:%M:%S %Z\n",
+                               time.gmtime()))
+    output.close()
     if os.path.exists(tarball + ".new"):
         os.rename(tarball + ".new", tarball)
 
@@ -428,7 +496,9 @@ def create_or_replace_chroot_tgz(config, tgz, tgz_ctrl, distro):
         statobj = os.stat(tgz)
         # stat.ST_MTIME is actually time file was initially created
         age = now - statobj[stat.ST_MTIME]
+        logging.info("Check-replace %s: age=%d vs. max=%d" % (tgz, age, max_tgz_age))
         if age > max_tgz_age:
+            logging.info("Limit-replace %s: last-retry=%d vs. min=%d" % (tgz, now - statobj[stat.ST_CTIME], min_tgz_retry_delay))
             # stat.ST_CTIME is time created OR last renamed
             if min_tgz_retry_delay is None or now - statobj[stat.ST_CTIME] > min_tgz_retry_delay:
                 os.rename(tgz, tgz + ".old")
@@ -448,8 +518,9 @@ def fetch_packages_file(config, distro):
     arch = config["arch"]
     if not arch:
         # Try to figure it out ourselves, using dpkg
-        vin, vout = os.popen2("dpkg --print-architecture")
-        arch = vout.read().rstrip()
+        p = subprocess.Popen(["dpkg", "--print-architecture"], 
+                             stdout=subprocess.PIPE)
+        arch = p.stdout.read().rstrip()
     packages_url = \
         "%s/dists/%s/main/binary-%s/Packages.bz2" % (mirror, distro, arch)
 
@@ -457,7 +528,7 @@ def fetch_packages_file(config, distro):
     f = piupartslib.open_packages_url(packages_url)
     packages_file = piupartslib.packagesdb.PackagesFile(f)
     f.close()
-    
+
     return packages_file
 
 
@@ -469,7 +540,7 @@ def create_file(filename, contents):
 
 def main():
     setup_logging(logging.INFO, None)
-    
+
     # For supporting multiple architectures and suites, we take a command-line
     # argument referring to a section in configuration file.  
     # If no argument is given, the "global" section is assumed.
@@ -488,19 +559,25 @@ def main():
                master_host=global_config["master-host"],
                master_user=global_config["master-user"],
                master_directory=global_config["master-directory"],
-               base_tgz_ctrl=[global_config["max-tgz-age"],global_config["min-tgz-retry-delay"]])
+               base_tgz_ctrl=[int(global_config["max-tgz-age"]),
+                              int(global_config["min-tgz-retry-delay"])])
         sections.append(section)
 
     while True:
-        for section in sections:
-            section.run()
-        idle = True
-        for section_name in section_names:
-          if os.listdir(os.path.join(global_config["master-directory"],section_name,"reserved")):
-            idle = False
-        if idle:
-          logging.info("Nothing to do, sleeping for %s seconds." % global_config["idle-sleep"])
-          time.sleep(int(global_config["idle-sleep"]))
+        test_count = 0
+        precedence = None
+
+        for section in sorted(sections, key=lambda section: section.precedence()):
+            if precedence is None or section.precedence() <= precedence:
+                processed = section.run()
+                if processed > 0:
+                    test_count += processed
+                    precedence = section.precedence()
+
+        if test_count == 0:
+            sleep_time = int(global_config["idle-sleep"])
+            logging.info("Nothing to do, sleeping for %d seconds." % sleep_time)
+            time.sleep( sleep_time )
 
 
 if __name__ == "__main__":
@@ -510,3 +587,5 @@ if __name__ == "__main__":
      print ''
      print 'Slave interrupted by the user, exiting... manual cleanup still neccessary.'
      sys.exit(1)  
+
+# vi:set et ts=4 sw=4 :
